import copy
from io import BytesIO
from typing import Dict, Optional, Sequence, Union

from .cookie_jar import CookieJar
from .headers import Headers
from .http_status import HttpStatus


class HttpResponse:
    def __init__(
        self,
        status: Union[int, HttpStatus] = HttpStatus.OK,
        body: Union[bytes, bytearray, str, None] = None,
        encoding: str = "utf-8",
        headers: Optional[Union[Dict[str, Union[str, Sequence[str]]], Headers]] = None,
    ):
        self._headers = headers if isinstance(headers, Headers) else Headers(headers)
        self.status_code = status
        self.body = BytesIO()
        self.encoding = encoding
        self.cookies = CookieJar()

        if body:
            self.write(body)

    @property
    def headers(self) -> Headers:
        headers: Headers = copy.copy(self._headers)
        for cookie in self.cookies.values():
            headers.set("Set-Cookie", cookie.serialise())
        return headers

    def write(self, body: Union[str, bytes, bytearray]) -> None:
        if isinstance(body, str):
            self.body.write(body.encode(self.encoding))
        else:
            self.body.write(body)

    @property
    def writable(self) -> bool:
        return not self.body.closed

    def close(self) -> None:
        self.body.close()

    def __str__(self) -> str:
        self.body.seek(0)
        return self.body.read().decode(self.encoding)


__all__ = ["HttpResponse"]



from typing import Any, Callable, Dict, Union

from .http_method import HttpMethod
from .http_request import HttpRequest
from .middleware import Middleware, MiddlewarePipeline, RoutingMiddleware
from .routing.route import Route


class Application:
    def __init__(self, *middleware: Union[Middleware, Callable]):
        self.middleware = MiddlewarePipeline()
        for item in middleware:
            self.middleware.append(item)

    def __call__(self, env: Dict[str, Any], start: Callable) -> Any:
        request = HttpRequest.from_wsgi(env)
        response = self.middleware(request)

        start(
            str(response.status_code),
            [(key, value) for key, value in response.headers.items()],
        )
        response.body.seek(0)
        return response.body


class ApplicationRouter(RoutingMiddleware):
    def get(self, route: str, **patterns) -> Callable:
        def _get(handler: Callable) -> Callable:
            self.methods[HttpMethod.GET].append(Route(route, **patterns), handler)
            return handler

        return _get

    def post(self, route: str, **patterns) -> Callable:
        def _post(handler: Callable) -> Callable:
            self.methods[HttpMethod.POST].append(Route(route, **patterns), handler)
            return handler

        return _post

    def put(self, route: str, **patterns) -> Callable:
        def _put(handler: Callable) -> Callable:
            self.methods[HttpMethod.PUT].append(Route(route, **patterns), handler)
            return handler

        return _put

    def patch(self, route: str, **patterns) -> Callable:
        def _patch(handler: Callable) -> Callable:
            self.methods[HttpMethod.PATCH].append(Route(route, **patterns), handler)
            return handler

        return _patch

    def delete(self, route: str, **patterns) -> Callable:
        def _delete(handler: Callable) -> Callable:
            self.methods[HttpMethod.DELETE].append(Route(route, **patterns), handler)
            return handler

        return _delete

    def head(self, route: str, **patterns) -> Callable:
        def _head(handler: Callable) -> Callable:
            self.methods[HttpMethod.HEAD].append(Route(route, **patterns), handler)
            return handler

        return _head

    def options(self, route: str, **patterns) -> Callable:
        def _options(handler: Callable) -> Callable:
            self.methods[HttpMethod.OPTIONS].append(Route(route, **patterns), handler)
            return handler

        return _options


router = ApplicationRouter()


def serve(*middleware, host: str = "0.0.0.0", port: int = 80) -> None:
    if not middleware:
        application = Application(router)
    else:
        application = Application(*middleware)

    try:
        from bjoern import run

        run(application, host=host, port=port)
    except ImportError as error:
        raise RuntimeError(
            "`bjoern` module could not be found, please run `pip install bjoern`."
        )


__all__ = ["Application", "router", "serve"]



import re
from copy import copy
from typing import Any, Dict, List, Optional, Pattern, Union

_ROUTE_REGEX = r"\\\{\s*(?P<var>[a-z_][a-z0-9_-]*)\s*\\\}"
_VAR_REGEX = "[^/]+"


class Route:
    def __init__(self, route: str, **kwargs):
        self.route = route
        self._part_patterns = kwargs
        self._attribute_names: List[str] = []
        self._pattern: Pattern[str] = None  # type: ignore
        self._attributes: Dict[str, str] = {}
        self.wildcard: bool = "*" in route

    @property
    def pattern(self) -> Pattern[str]:
        if not self._pattern:
            self._parse()
        return self._pattern

    def _parse(self) -> None:
        def _parse_var(match):
            attribute = match.group(1)
            self._attribute_names.append(attribute)
            if attribute in self._part_patterns:
                return f"({self._part_patterns[attribute]})"
            return f"({_VAR_REGEX})"

        pattern = re.escape(self.route)
        pattern = re.sub(_ROUTE_REGEX, _parse_var, pattern, re.I | re.M)
        pattern = re.sub(r"\\\*", ".*?", pattern)
        self._pattern = re.compile("^" + pattern + "$", re.I | re.M,)

    def match(self, uri: str) -> Union[bool, "Route"]:
        matches = self.pattern.findall(uri)
        if not matches:
            return False

        if not self._attribute_names:
            return copy(self)

        if isinstance(matches[0], tuple):
            matches = list(matches[0])

        route = copy(self)
        match_index = 0
        for value in matches:
            route._attributes[self._attribute_names[match_index]] = value
            match_index += 1

        return route

    @property
    def attributes(self):
        return self._attributes

    def __str__(self) -> str:
        return self.route

    def __bool__(self) -> bool:
        return True

    def __getitem__(self, key: str) -> str:
        return self._attributes[key]

    def __contains__(self, key: str) -> bool:
        return key in self._attributes

    def get(self, key: str, default: Optional[Any] = None):
        if key in self:
            return self[key]

        return default


__all__ = ["Route"]



class HttpError(Exception):
    status_code: int = 500
    http_message = "Internal Server Error"

    def __init__(self, http_message: str, status_code: int = None):
        if status_code:
            self.status_code = status_code
        if http_message:
            self.http_message = http_message

        super().__init__(http_message)


__all__ = ["HttpError"]



from .chocs import Application, router, serve
from .cookie_jar import Cookie, CookieJar, CookieSameSitePolicy
from .headers import Headers
from .http_method import HttpMethod
from .http_request import HttpRequest
from .http_response import HttpResponse
from .http_status import HttpStatus
from .query_string import QueryString, parse_qs



from cgi import parse_header
from enum import Enum
from io import BytesIO
from tempfile import TemporaryFile
from typing import Any, Dict, IO, Tuple

from .form_body import FormBody


class ParserState(Enum):
    PART_BOUNDARY = 0
    CONTENT_DISPOSITION = 1
    CONTENT_TYPE = 2
    CONTENT_HEADER = 3
    CONTENT_DATA = 4
    END = 5


class MultipartBody(FormBody):
    @staticmethod
    def from_wsgi(
        wsgi_input: BytesIO, encoding: str = "utf8", boundary: str = ""
    ) -> "MultipartBody":
        assert boundary, (
            "%s.from_wsgi requires boundary parameters." % MultipartBody.__name__
        )
        wsgi_input.seek(0)
        return _parse_multipart_data(wsgi_input.read(), boundary, encoding)


def _parse_multipart_data(
    data: bytes, boundary: str, encoding: str = "utf8"
) -> MultipartBody:
    state = ParserState.PART_BOUNDARY
    prev_byte = None
    cursor = 0
    boundary_length = len(boundary)
    string_buffer = ""
    body = MultipartBody()

    def _append_content_to_body(
        raw_content_disposition: str, _content_type: str, _content_data: bytes
    ) -> None:
        parsed_content_disposition: Tuple[str, Dict[str, str]] = parse_header(
            raw_content_disposition[20:]
        )
        if "filename" in parsed_content_disposition[1]:
            tmp_file = TemporaryFile()
            tmp_file.write(_content_data)
            tmp_file.seek(0)
            body[parsed_content_disposition[1]["name"]] = UploadedFile(
                tmp_file,
                _content_type[14:].lower(),
                parsed_content_disposition[1]["filename"],
            )
        else:
            body[parsed_content_disposition[1]["name"]] = _content_data.decode(encoding)

    content_disposition = ""
    content_type = ""
    content_cursor = 0

    for code in data:
        line_break = code == 0x0A and prev_byte == 0x0D
        new_line_char = code == 0x0A or code == 0x0D

        if not new_line_char:
            string_buffer = string_buffer + chr(code)

        if state is ParserState.PART_BOUNDARY and line_break:
            if string_buffer == "--" + boundary:
                string_buffer = ""
                state = ParserState.CONTENT_DISPOSITION
            else:
                raise IOError(
                    "Could not parse message body, body is malformed or incorrect boundary was passed."
                )

        elif state is ParserState.CONTENT_DISPOSITION and line_break:
            content_disposition = string_buffer
            string_buffer = ""
            state = ParserState.CONTENT_TYPE
            if (
                data[cursor + 1 : cursor + 13].decode(encoding).lower()
                != "content-type"
            ):
                state = ParserState.CONTENT_HEADER
        elif state is ParserState.CONTENT_TYPE and line_break:
            content_type = string_buffer
            state = ParserState.CONTENT_HEADER
            string_buffer = ""
        elif state is ParserState.CONTENT_HEADER and line_break:
            string_buffer = ""
            state = ParserState.CONTENT_DATA
            content_cursor = cursor
        if state is ParserState.CONTENT_DATA:
            if line_break and string_buffer == "--" + boundary:
                content_data = data[content_cursor + 1 : cursor - (boundary_length + 5)]
                _append_content_to_body(content_disposition, content_type, content_data)
                content_type = ""
                state = ParserState.CONTENT_DISPOSITION

            if line_break and string_buffer == "--" + boundary + "--":
                content_data = data[content_cursor + 1 : cursor - (boundary_length + 7)]
                _append_content_to_body(content_disposition, content_type, content_data)
                content_type = ""
                state = ParserState.END

            if line_break:
                string_buffer = ""

        prev_byte = code
        cursor += 1

    return body


class UploadedFile:
    """
    Proxy class for uploaded file
    """

    def __init__(self, file: IO[Any], mimetype: str, filename: str):
        self.file = file
        self.mimetype = mimetype
        self.filename = filename
        self.length = 0
        self._str = ""

    def read(self) -> bytes:
        return self.file.read()

    def seek(self, offset: int) -> int:
        return self.file.seek(offset)

    def close(self) -> None:
        self.file.close()

    def save(self, path: str) -> IO[Any]:
        if self.file.closed:
            raise ValueError(f"Cannot save to file {path} of closed stream.")
        with open(path, "wb") as file:
            self.seek(0)
            file.write(self.read())

        return file

    def __float__(self) -> None:
        raise ValueError(
            f"Cannot convert instance of {TemporaryFile.__name__} to float"
        )

    def __int__(self) -> None:
        raise ValueError(f"Cannot convert instance of {TemporaryFile.__name__} to int")

    def __len__(self) -> int:
        if not self.length:
            bytes = self.read()
            self.seek(0)
            self.length = len(bytes)
        return self.length

    def __bool__(self) -> bool:
        return len(self) > 0

    def __str__(self) -> str:

        if not self._str:
            self._str = self.read().decode()

        return self._str

    def __enter__(self) -> IO[Any]:
        return self.file

    def __exit__(self, *args: Any) -> None:
        self.close()


__all__ = ["MultipartBody", "UploadedFile"]



from cgi import parse_header
from io import BytesIO
from typing import Any, Dict, Optional, Tuple, Union

from .cookie_jar import CookieJar, parse_cookie_header
from .headers import Headers
from .http_method import HttpMethod
from .message.body import RequestBody
from .message.form_body import FormBody
from .message.json_body import JsonBody
from .message.multipart_body import MultipartBody
from .query_string import QueryString


class HttpRequest:
    def __init__(
        self,
        method: HttpMethod,
        uri: str = "/",
        body: Optional[BytesIO] = None,
        query_string: Optional[QueryString] = None,
        headers: Optional[Headers] = None,
    ):
        self.headers = headers if headers else Headers()
        self.body = body if body else BytesIO(b"")
        self.method = method
        self.uri = uri
        self.query_string = query_string
        self._parsed_body: Union[RequestBody, str] = ""
        self.attributes: Dict[str, str] = {}
        self._cookies: Optional[CookieJar] = None

    @property
    def parsed_body(self) -> Union[RequestBody, str]:
        if self._parsed_body:
            return self._parsed_body

        content_type: Tuple[str, Dict[str, str]] = parse_header(
            self.headers.get("Content-Type")  # type: ignore
        )

        if content_type[0] == "multipart/form-data":
            body: Union[RequestBody, str] = MultipartBody.from_wsgi(
                self.body,
                content_type[1].get("charset", ""),
                content_type[1].get("boundary", ""),
            )
        elif content_type[0] == "application/x-www-form-urlencoded":
            body = FormBody.from_wsgi(self.body, content_type[1].get("charset", "utf8"))

        elif content_type[0] == "application/json":
            body = JsonBody.from_wsgi(self.body, content_type[1].get("charset", "utf8"))
        else:
            self.body.seek(0)
            body = self.body.read().decode(content_type[1].get("charset", "utf8"))

        self._parsed_body = body

        return self._parsed_body

    @classmethod
    def from_wsgi(cls, environ: Dict[str, Any]) -> "HttpRequest":
        headers = Headers()
        for key, value in environ.items():
            if not key.startswith("HTTP"):
                continue
            headers.set(key, value)
        headers.set("Content-Type", environ.get("CONTENT_TYPE", "text/plain"))
        return cls(
            method=HttpMethod(environ.get("REQUEST_METHOD", "GET").upper()),
            uri=environ.get("PATH_INFO", "/"),
            body=environ.get("wsgi.input", BytesIO(b"")),
            query_string=QueryString(environ.get("QUERY_STRING", "")),
            headers=headers,
        )

    @property
    def cookies(self):
        if self._cookies is None:
            self._cookies = parse_cookie_header(self.headers.get("cookie"))

        return self._cookies


__all__ = ["HttpRequest"]



from abc import ABC, abstractmethod

from chocs.http_request import HttpRequest
from chocs.http_response import HttpResponse


class MiddlewareHandler(ABC):
    @abstractmethod
    def __call__(self, request: HttpRequest) -> HttpResponse:
        pass


__all__ = ["MiddlewareHandler"]



import re
from datetime import datetime
from enum import Enum
from typing import Dict, ItemsView, KeysView, Optional, Union, ValuesView
from urllib.parse import quote, unquote

from chocs.errors.cookie_error import CookieError

COOKIE_NAME_VALIDATOR = re.compile(r"[a-z0-9!#$%&'*+.^_`|~\-]+", re.I)


class CookieSameSitePolicy(Enum):
    STRICT = "Strict"
    LAX = "Lax"
    NONE = "None"


class Cookie:
    def __init__(
        self,
        name: str,
        value: str,
        path: Optional[str] = None,
        domain: Optional[str] = None,
        expires: Optional[datetime] = None,
        max_age: Optional[int] = None,
        secure: bool = False,
        http_only: bool = False,
        same_site: Union[bool, CookieSameSitePolicy] = False,
    ):
        if not COOKIE_NAME_VALIDATOR.match(name):
            raise CookieError(
                f"Invalid cookie name {name}, cookie name must be valid RFC 2616 token."
            )
        self._name: str = name
        self.value = value
        self.path = path
        self.domain = domain
        self.expires = expires
        self.max_age = max_age
        self.secure = secure
        self.http_only = http_only
        self.same_site = same_site

    @property
    def name(self):
        """
       Cookie's name must be valid RFC-2616 token. Once set should not be changed as this causes some implications
       how we keep cookies in the cookie jar.
       .. _RFC-2616 token: https://tools.ietf.org/html/rfc2616#section-2.2
       """
        return self._name

    @property
    def safe_value(self) -> str:
        return quote(self.value)

    def __str__(self) -> str:
        return self.value

    def __int__(self) -> int:
        return int(self.value)

    def __float__(self) -> float:
        return float(self.value)

    def __bool__(self) -> bool:
        return bool(self.value)

    def serialise(self) -> str:
        output = f"{self.name}={self.safe_value}"
        if self.max_age is not None:
            output += f"; Max-Age={self.max_age}"

        if self.domain is not None:
            output += f"; Domain={self.domain}"

        if self.path is not None:
            output += f"; Path={self.path}"

        if self.expires is not None:
            output += f"; Expires={self.expires.strftime('%a, %d %b %Y %H:%M:%S %z')}"

        if self.secure:
            output += "; Secure"

        if self.http_only:
            output += "; HttpOnly"

        if self.same_site:
            if isinstance(self.same_site, CookieSameSitePolicy):
                output += f"; SameSite={self.same_site.value}"
            else:
                output += "; SameSite=Strict"

        return output


class CookieJar:
    def __init__(self):
        self._cookies: Dict[str, Cookie] = {}

    def append(self, cookie: Cookie):
        self._cookies[cookie.name] = cookie

    def __setitem__(self, key: str, value: str):
        if isinstance(value, str):
            cookie = Cookie(key, value)
        else:
            raise CookieError(
                "`chocs.cookie_jar.CookieJar.__setitem__` accepts only string values. "
                "To append new cookie use `chocs.cookie_jar.CookieJar.append` method instead."
            )
        self._cookies[key] = cookie

    def __getitem__(self, key: str) -> Cookie:
        return self._cookies[key]

    def __delitem__(self, key):
        del self._cookies[key]

    def __contains__(self, key) -> bool:
        return key in self._cookies

    def __len__(self) -> int:
        return len(self._cookies)

    def items(self) -> ItemsView[str, Cookie]:
        return self._cookies.items()

    def values(self) -> ValuesView[Cookie]:
        return self._cookies.values()

    def keys(self) -> KeysView[str]:
        return self._cookies.keys()


def parse_cookie_header(header: str) -> CookieJar:
    """
    When the user agent generates an HTTP request, the user agent MUST
    NOT attach more than one Cookie header field.
    https://tools.ietf.org/html/rfc6265#section-5.4

    Therefore parse_cookie_header will only accept a single header string
    """
    result = CookieJar()

    cookies = header.split(";")
    for cookie in cookies:
        separator_position = cookie.find("=")
        if separator_position < 0:  # Broken cookie?
            continue
        try:
            result.append(
                Cookie(
                    cookie[0:separator_position].strip(),
                    unquote(cookie[separator_position + 1 :].strip()),
                )
            )
        except CookieError:  # Invalid cookie name
            continue
    return result


__all__ = ["Cookie", "CookieJar", "parse_cookie_header", "CookieSameSitePolicy"]
